package Impedance::Base;
#$Id: Base.pm,v 1.24 2013/05/06 10:28:55 sqko Exp $
#################################################################
#
#
#      Impedance::Base
#
# This module defines all member functions which an 
# ElchemeaAnalytical Impedance element must honor.
# Most of the functions are merely stups intended to be 
# overloaded by the individual class definitions.
#
# The module also includes class defininitions on the 
# basic discrete elements (R, C, L and Q, Q beeing the constant
# phase element).
#
# The module is intended to be used in conjunction with gnuplot
# but can be used as is (but no fitting will be possible).
#
# To obtain a element instance call one of the constructors 
# as show below:
#
# $id = Impedance::Base->new($);
#
# $id = Impedance::R->new($);
#
# $id = Impedance::C->new($);
#
# $id = Impedance::L->new($);
#
# $id = Impedance::Q->new($);
#
# All the constructors accepts a single argument which must be 
# the impedance element number. It is advisable to make sure
# that element id's are unique in order to be able to distinguish.
# 
# All Impedance element instances has the following member 
# functions:
# 
# $id->type()            Returns the type of impedance element. 
# $id->name()            Returns the name of the element 
#                        (usually the type and the id)
# $id->nr()              Returns the element id.
# $id->function_name()   Returns the Gnuplot function name of the 
#                        element.
# $id->value($[$])       Sets or gets the value of the specified element 
#                        parameter (If 2 arguments are specified, the 
#                        second is the value to be set).
# $id->description($)    Returns the description for the specified tag.
#                        valid tags are 'name' and the tags returned by
#                        the tags() member function. The 'name' tag returns
#                        a text string describing the whole element
#                        and the other strings descriobe the specified
#                        parameter.
#                        If called with no arguments, it returns the 
#                        string for the 'name' tag.
# $id->tags()            Return a list of valid element parameter names.
# $id->functions()       Returns a string containing all the functions 
#                        and variable decleratins for gnuplot.
# $id->F($)              Returns the impedance value of the element at
#                        the specified frequency. The returned impedance
#                        is of type Math::Complex.
# $id->fit()             Returns a list of possible fit variables 
#                        (all starting with value 1) for use with gnuplot. 
# $id->save()            Returns a string containing functions to save the 
#                        final fit variables (base value multiplied with 
#                        fit variable). This is also for gnuplot use.
# $id->print_line()      Returns a string defining the impedance element.
#                        The format is described below.
# $id->helperfunctions() Returns a list of paramter names for which 
#                        special help functions exist for calculating 
#                        usefull estimates for start values for fitting.
#                        Note, that some impedance elements does not
#                        have any way of determining good start values
#                        in which case the helperfunctions() member function
#                        simply returns an empty list. 
# $id->helpfunction(@)   This function has multiple uses. If called 
#                        without any arguments, it returns the list
#                        from the helperfunctions() member function.
#                        If called with one argument it returns a string
#                        describing how many additional arguments must
#                        be parsed to it (which may wary depending on 
#                        impedance element) and in which order.
#                        If called with more than one argument, the first
#                        Argument is the parameter name to calculate start
#                        value for, and the additional arguments (which must be
#                        a string of the form: 
#                        "$frequency $real_value $imaginary_value")
#                        are used for the calculation.
# $id->f_imax()          Returns the frequency for which the imaginary
#                        value of the impedance is at it's maximum (negative)
#                        value. If the element type makes such a calculation
#                        invalid, the function returns undefined.
#                        If the frequency is either infinite or the DC case, 
#                        the reported frequencies will be 1e100 Hz and 1e-100 Hz 
#                        respectively.
#                        For frequency independent elements the function returns 0.
#
# The print_line member function returns a string which can be used to reload
# an impedance element. The format is:
#
# Type: parameter1=value1, parameter2=value2, ....
#
# An exception to this is the container elements (serial and parallel)
# which is defined and implemented in the Impedance::Complex class.
#
#
############################################################################

use strict;
use warnings;
use Math::Complex;
use Debug;
our $AUTOLOAD;
our @ISA = qw(Debug);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new
{
    my $class = shift;
    my $nr = shift;
    die "Error: No number specified as argument to new impedance element!\n" unless ($nr && $nr =~ /^\d+$/);
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Default';
    $self->{'_nr'} = $nr;
    $self->{'_values'} = {};
    $self->{'_tags'} = [];
    $self->{'_helptags'} = [];
    $self->{'_desc'} = {};
    $self->{'_f_imax'} = undef;
    bless $self,$class;
    return $self;
}


sub f_imax()
{
    my $self = shift;
    return $self->{'_f_imax'};
}

sub type()
{
    my $self = shift;
    return $self->{'_name'};
}

sub nr()
{
    my $self = shift;
    return $self->{'_nr'};

}

sub description
{
    my $self = shift;
    my $key = shift;
    return $self->{'_desc'}{$key} if ($key && exists($self->{'_desc'}{$key}));
    return $self->{'_desc'}{'name'};
}

sub value
{
    my $self = shift;
    my $key = shift;
    if (@_)
    {
	my $val = shift;
	die "Error: No key and/or value supplied!" unless ($key && defined($val));
	$self->{'_values'}{$key} = $val + 0;
    }
    return $self->{'_values'}{$key} if exists($self->{'_values'}{$key});
    return '1.0';
}

sub F($)
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(0,0);
}

sub name()
{
    my $self = shift;
    return $self->{'_name'}.'_'.$self->{'_nr'};
}

sub function_name()
{
    my $self = shift;
    return 'z_'.$self->{'_nr'}.'(x)';
}

sub tags()
#### Returns a list of parameterw whose start values must be set ###
{
    my $self = shift;
    return @{$self->{'_tags'}};
}

sub fit()
#### Returns a list of possible fit variables (all starting with value 1) ###
{
    my $self = shift;
    die "Error: ".$self->{'_name'}." has not defined the fit() member function!\n";
}

sub functions()
#### Returns a string cojntaining all the functions and varuiable declarations for gnuplot ###
{
    my $self = shift;
    die "Error: ".$self->{'_name'}." has not defined the functions() member function!\n";
}

sub save()
### Returns a string containing functions to save the final fit values ###
### (base value multiply with fit variable)                            ###
### Overload for some complex elements (notably RQ's)                  ###
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ();
    foreach (@{$self->{'_tags'}})
    {
	push @l,"Final_".$_."_$n = ".$_."_$n * f".lc($_)."_$n";
    }
    return join("\n",@l)."\n";
}

sub helperfunctions()
{
    my $self = shift;
    return @{$self->{'_helptags'}};			
}	

sub helpfunction()
{
    ### Overload vhere apropriate ###
    my $self = shift;
    unless (@_)
    {
	return $self->helperfunctions();
    }
    my $tag = shift;
    return undef;
}


sub print_line()
{
    my $self = shift;
    my $res = $self->type().': ';
    my @l = ();
    foreach my $t ($self->tags)
    {
	push @l,$t.'='.$self->value($t);
    }
    return $res.join(', ',@l);
}

sub AUTOLOAD()
{
    ################################################################
    ### AUTOLOAD()                                               ###
    ### If caller ended here, undefiend subroutine call happened ###
    ################################################################
    my $self = shift;
    my $type = ref ($self) or die "$self is not an object!";
    die "Unknown function call $AUTOLOAD attempted!";
}

sub DESTROY()
{

}


package Impedance::L;
use Math::Complex;
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'L';
    $self->{'_tags'} = ['L'];
    $self->{'_values'}{'L'} = 1e-9; ## Default 1 nH ##
    $self->{'_desc'}{'L'} = 'Inductance in Henry';
    $self->{'_desc'}{'name'} = 'Inductor';
    $self->{'_f_imax'} = 1e100;
    bless $self,$class;
    return $self;
}

sub F()
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return $self->{'_values'}{'L'} * cplx(0,1) * 2 * pi() *$f;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("L_$n = ".$self->value('L'),"fl_$n = 1.0");
    push @l,'z_'.$n.'(x)=L_'.$n.'*fl_'.$n.'*{0,1}*2*pi*x';
    return join("\n",@l)."\n";
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fl_$n");
}

package Impedance::R;
use Math::Complex;
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'R';
    $self->{'_tags'} = ['R'];
    $self->{'_desc'}{'name'} = 'Resistor';
    $self->{'_desc'}{'R'} = 'Resistance in Ohm';
    $self->{'_f_imax'} = 0;
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("R_$n = ".$self->value('R'),"fr_$n = 1.0");
    push @l,'z_'.$n.'(x)=R_'.$n.'*fr_'.$n;
    return join("\n",@l)."\n";
}

sub F()
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx($self->value('R'),0);
}

sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fr_$n");
}

package Impedance::Q;
use Math::Complex;
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'Q';
    $self->{'_tags'} = ['Y','n'];
    $self->{'_values'}{'Y'} = 1e-4;  ## Default 100 'yF' ##
    $self->{'_values'}{'n'} = 0.8; 
    $self->{'_desc'}{'name'} = 'Constant phase element';
    $self->{'_desc'}{'Y'} = "'Capacitance' in F*s^(n-1)";
    $self->{'_desc'}{'n'} = "phase angle in '1/2 pi' radians (1 means 90 degrees)";
    $self->{'_f_imax'} = 1e-100;
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("Y_$n = ".$self->value('Y'),"fy_$n = 1.0","n_$n = ".$self->value('n'));
    push @l,'z_'.$n.'(x)=1.0/(Y_'.$n.'*fy_'.$n.'*({0,1}*2*pi*x)**n_'.$n.')';
    return join("\n",@l)."\n";
}

sub F()
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return cplx(1,0) / ($self->value('Y') * (cplx(0,1) * 2 * pi() * $f)**$self->value('n'));
}


sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fy_$n","n_$n");
}

package Impedance::C;
use Math::Complex;
our @ISA = qw(Impedance::Base);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_name'} = 'C';
    $self->{'_tags'} = ['C'];
    $self->{'_values'}{'C'} = 1e-4;  ## Default 100 yF ##
    $self->{'_desc'}{'C'} = "Capacitance in Farad";
    $self->{'_desc'}{'name'} = 'Capacitor';
    $self->{'_f_imax'} = 1e-100;
    bless $self,$class;
    return $self;
}

sub functions()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    my @l = ("C_$n = ".$self->value('C'),"fc_$n = 1.0");
    push @l,'z_'.$n.'(x)=-{0,1}/(C_'.$n.'*fc_'.$n.'*2*pi*x)';
    return join("\n",@l)."\n";
}

sub F()
{
    my $self = shift;
    my $f = shift;
    return undef unless ($f && $f > 0);
    return -cplx(0,1) / ($self->value('C') * 2 * pi() * $f);
}


sub fit()
{
    my $self = shift;
    my $n = $self->{'_nr'};
    return ("fc_$n");
}

return 1;
